#pragma once

#include <cmath>
#include <vector>
#include "math/matrix.hh"

/*
  computation of the coefficients h_{k1,k2} of the expansion of the impact factor

  (1)   h_Higgs(M1,M2) = sum_{k1,k2>=0} h_{k1,k2} M1^k1 M2^k2

  where M1 and M2 are the Mellin conjugate variables of the two gluon off-shellness.
*/

namespace HELL {

class HiggsSmallX {
public:
  HiggsSmallX() { init(125, 173); };
  HiggsSmallX(double mH, double mt, double mb=-1, double mc=-1, double muFrat=1) { init(mH,mt,mb,mc,muFrat); };
  ~HiggsSmallX() { return; };
  void   SetFactorizationScale(double muFrat) { _muFrat=muFrat; };
  double GetFactorizationScale() { return _muFrat; };
  void init(double mH, double mt, double mb=-1, double mc=-1, double muFrat=1);
  // this function returns a vector A such that h_{k1,k2} = A[k1+k2][k1], up to k1+k2 <= maxorder
  std::vector<std::vector<double> > hCoeffs(int maxorder=3);
  std::vector<std::vector<double> > hCoeffsSaved();
  void initCoeffs(int maxorder=3);
  double h1();
  double factorH();
  double h_kp_km(int kp, int km, double *err = NULL);
  //
  double sf2    (double t, double y);
  double sf1    (double t, double y);
  double sf1dt  (double t, double y);
  double sf1dy  (double t, double y);
  double sf1dtdy(double t, double y);
  double sf1y1dt(double t);
  //
private:
  double prefH;
  double _muFrat;
  bool areCoeffsSaved;
  std::vector<std::vector<double> > savedCoeffs;
  std::vector<dcomplex> Yi;
};

};

