#include <iomanip>
#include <iostream>
#include <fstream>
#include <string>
#include <sstream>
#include <cmath>
#include <cstdlib>
//#include <cuba.h>
#include <gsl/gsl_errno.h>

#include "CoefficientFunctions.hh"
#include "CoefficientFunctionsOffShell.hh"
#include "CoefficientFunctionsOffShellInterpolator.hh"
#include "hHiggs_new_var.hh"
#include "Pqg.hh"
#include "math/borel.hh"
#include "integration.hh"
//#include "gammaNLO.hh"
#include "h_coefficients.hh"
//#include <complex_bessel/sp_airy.h>


using namespace std;



namespace HELL {


  unsigned int RCmode = 0;
  void SetRCmodeCF(unsigned int mode) {  // mode=0 uses gammadot, mode=1 uses as*beta0
    if(mode>1) {
      cout << "Error: invalid RC mode. Using default." << endl;
      mode = 0;
    }
    RCmode = mode;
  }
  bool RCvar = 0;
  void SetRCvarCF(bool var) {  // var=true uses "Aiy" anomalous dimension
    RCvar = var;
  }



  double myArg(dcomplex x) {
    double ar = atan2(imag(x),real(x));
    //if(ar<-M_PI/2) ar += 2*M_PI;
    return ar;
  }
  dcomplex myLog(dcomplex z) {
    return log(abs(z)) + I*myArg(z);
  }
  dcomplex evolU(double xi, dcomplex g, dcomplex r) {
    if(xi==0) return 0.;
    if(xi==1) return 1.;
    dcomplex res = exp( g/r * myLog(r*log(xi)+1.) );
    //if(imag(r)<0 && real(r)*log(xi)+1.<0) res *= exp( g/r * I*2.*M_PI);
    //if(imag(r)<0 && real(r)*log(xi)+1.<0) res = exp( g/r * (log(r*log(xi)+1.)+I*2.*M_PI) );
    return res;
  }
  dcomplex Dxi_evolU(double xi, dcomplex g, dcomplex r) { // derivative wrt xi
    if(xi==0) return 0.;
    if(xi==1) return g;
    dcomplex res = exp( (g/r-1.) * myLog(r*log(xi)+1.) ) * g/xi;
    return res;
  }
  dcomplex Dxi_log_evolU(double xi, dcomplex g, dcomplex r) { // derivative wrt xi
    dcomplex res = g/xi/(r*log(xi)+1.);
    return res;
  }


  dcomplex integrand(double xi, dcomplex N, dcomplex gamma, dcomplex r, process p, double asb0, double x1, void *par) {
    double xiU = xi;
    double xi0 = exp(-1./asb0);
    if(p == ggHaux) {
      HiggsSmallX *ggHsx = (HiggsSmallX*) par;
      double muFrat = ggHsx->GetFactorizationScale();
      xiU /= muFrat*muFrat;
    }
    if(xiU<xi0) {
      if(fabs(xiU-xi0)/xi0<1e-3) {
	xiU = xi0;
      } else {
	cout << "cazzo!  xi0=" << xi0 << "  xiU-xi0=" << xiU-xi0 << "   xi="<< xi << endl;
	return 0;
      }
    }
    dcomplex  U =     evolU(xiU,gamma,r);
    dcomplex dU = Dxi_evolU(xiU,gamma,r);
    if(xiU<1) {  // apply higher-twist damping
      int kk = 1;
      double  HTdamp = 1 - pow(-asb0*log(xiU),kk*(1+1/asb0));
      double dHTdamp = (HTdamp-1) * kk*(1+1/asb0)/xiU/log(xiU);
      dU = dU * HTdamp + U * dHTdamp;
      U *= HTdamp;
    }
    dcomplex res = U;
    //
    if(p == gamma_qg) res *= - Dxi_Hqg(xi);
    //
    //if(p == CL) res *= - Dxi_CLoff(xi, x1);
    //if(p == C2) res *= - Dxi_C2off(xi, x1) - (xi<=1 ? 1./xi : 0);
    if(p == CL) res *= - Dxi_CLoff(xi);
    if(p == C2) res *= - Dxi_C2off(xi) - (xi<=1 ? 1./xi : 0);
    //if(p == CL) res = CLoff(xi) * dU;
    //if(p == C2) res = (C2off(xi) + (xi<=1 ? log(xi) : 0)) * dU;
    //
    //if(p == Ctest1) res = - (xi<=1 ? 1. : 0) * U + (xi<=1 && xi>=0.9 ? 10. : 0);
    //if(p == Ctest2) res = (xi<=1 ? xi : 0) * dU;
    //if(p == Ctest1) res = - Dxi_CLoff(xi) * U;
    //if(p == Ctest2) res = CLoff(xi) * dU;
    //
    if(p>=50 && p<60) {
      double mQ = *(double*) par;
      if(p == Khg) res *= - Dxi_Khgoff(xi,mQ) + (xi<=1 ? 1./xi : 0);
    }
    if(p >= 60 && p<100) {
      Interpolator *interp = (Interpolator*) par;
      if(p == MC2)   res *= - Dxi_MC2off(xi, interp);
      if(p == MCL)   res *= - Dxi_MCLoff(xi, interp);
      if(p == MC2ax) res *= - Dxi_MC2axoff(xi, interp);
      if(p == MCLax) res *= - Dxi_MCLaxoff(xi, interp);
      //
      double xim = interp->Getxim();
      if(p == MC2cc) res *= - Dxi_MC2CCoff(xi, interp) -               (xi<=1 ? 1./xi : 0) /2;
      if(p == MCLcc) res *= - Dxi_MCLCCoff(xi, interp) - xim/(1+xim) * (xi<=1 ? 1./xi : 0) /2;
      if(p == MC3cc) res *= - Dxi_MC3CCoff(xi, interp) -   1/(1+xim) * (xi<=1 ? 1./xi : 0) /2;
    }
    if(p == ggHaux) {
      HiggsSmallX *ggHsx = (HiggsSmallX*) par;
      res *= ggHsx->sf1y1dt(xi/2)/2;
    }
    if(p == DY) {
      res *= - Dxi_CDYoff(xi) - (xi<=1 ? 1./xi : 0);
    }
    //
    //cout << setw(15) << xi << setw(15) << x1 << setw(35) << res << endl;
    return res;
  }
  dcomplex integrand2d(double t, double y, dcomplex gamma, dcomplex r, process p, double asb0, void *par) {
    double xi1 = t*(1+y);
    double xi2 = t*(1-y);
    double xiU1 = xi1;
    double xiU2 = xi2;
    double xi0 = exp(-1./asb0);
    double muFrat = 1;
    if(p == ggH) {
      HiggsSmallX *ggHsx = (HiggsSmallX*) par;
      muFrat = ggHsx->GetFactorizationScale();
      xiU1 /= muFrat*muFrat;
      xiU2 /= muFrat*muFrat;
    }
    if(xiU1<xi0) {
      if(fabs(xiU1-xi0)/xi0<1e-3) {
	xiU1 = xi0;
      } else {
	cout << "cazzo!  xi0=" << xi0 << "  xi1-xi0=" << xiU1-xi0 << "   t="<<t << " y="<<y << endl;
	return 0;
      }
    }
    if(xiU2<xi0) {
      if(fabs(xiU2-xi0)/xi0<1e-3) {
	xiU2 = xi0;
      } else {
	cout << "cazzo!  xi0=" << xi0 << "  xi2-xi0=" << xiU2-xi0 << "   xi2="<<xiU2 << "   t="<<t << " y="<<y << endl;
	return 0;
      }
    }
    if(xiU1<xi0 || xiU2<xi0) return 0;
    dcomplex  U1 =              evolU(xiU1,gamma,r);
    //dcomplex dU1 = U1 * Dxi_log_evolU(xiU1,gamma,r);
    dcomplex dU1 = Dxi_evolU(xiU1,gamma,r);
    if(xiU1<1) {  // apply higher-twist damping
      int kk = 1;
      double  HTdamp = 1 - pow(-asb0*log(xiU1),kk*(1+1/asb0));
      double dHTdamp = (HTdamp-1) * kk*(1+1/asb0)/xiU2/log(xiU1);
      dU1 = dU1 * HTdamp + U1 * dHTdamp;
      U1 *= HTdamp;
    }
    dcomplex  U2 =              evolU(xiU2,gamma,r);
    //dcomplex dU2 = U2 * Dxi_log_evolU(xiU2,gamma,r);
    dcomplex dU2 = Dxi_evolU(xiU2,gamma,r);
    if(xiU2<1) {  // apply higher-twist damping
      int kk = 1;
      double  HTdamp = 1 - pow(-asb0*log(xiU2),kk*(1+1/asb0));
      double dHTdamp = (HTdamp-1) * kk*(1+1/asb0)/xiU2/log(xiU2);
      dU2 = dU2 * HTdamp + U2 * dHTdamp;
      U2 *= HTdamp;
    }
    //
    dU1 /= muFrat*muFrat;
    dU2 /= muFrat*muFrat;
    dcomplex res = 0.;
    //
    if(p == ggH) {
      HiggsSmallX *ggHsx = (HiggsSmallX*) par;
      //double muFrat = ggHsx->GetFactorizationScale();
      //xi1 *= muFrat*muFrat;
      //xi2 *= muFrat*muFrat;
      //t   *= muFrat*muFrat;
      //res = 4. * dU1 * dU2 * t * ( xi1*xi2*ggHsx->sf2(t,y) + ggHsx->sf1(t,y) );
      res = 4. * dU1 * dU2 * t * xi1*xi2*ggHsx->sf2(t,y) - U1 * U2 * ( 2*y*ggHsx->sf1dtdy(t,y) + (1.+xi1*dU1/U1+xi2*dU2/U2)*ggHsx->sf1dt(t,y) - (dU1/U1-dU2/U2)*(1-y*y)*ggHsx->sf1dy(t,y) );
      //res = 4. * dU1 * dU2 * t * xi1*xi2*ggHsx->sf2(t,y) - U1 * U2 * (   y*ggHsx->sf1dtdy(t,y) + (1.+xi1*dU1/U1+xi2*dU2/U2)*ggHsx->sf1dt(t,y) - (xi1*dU1/U1-xi2*dU2/U2)/t*ggHsx->sf1dy(t,y) );
      //res /= muFrat*muFrat;
    }
    //cout << res << endl;
    //
    return res;
  }
  struct s_pr3c2d { process p; dcomplex c1, c2, c3; double d3, d4; void *par; };
  int integrandCUBAgeneral(const int *ndim, const double x[], const int *ncomp, double res[], void *pars, int whichFunction) {
    //if(x[1]==0)
    s_pr3c2d p = *(s_pr3c2d*)pars;
    //dcomplex N  = p.c1;
    dcomplex g  = p.c2;
    dcomplex r  = p.c3;
    double a0 = p.d3;  // as(Q^2)
    double b0 = p.d4;  // beta0
    //
    void *par = p.par;
    process pr = p.p;
    double lFQ = 0;
    double xiF = 1;
    if(pr == ggH) {
      HiggsSmallX *ggHsx = (HiggsSmallX*) par;
      double muFrat = ggHsx->GetFactorizationScale();
      lFQ = 2*log(muFrat);
      xiF = muFrat*muFrat;
    }
    dcomplex result = 0;
    //
    if(whichFunction == 0) {
      // y change of variable: yold = 1-y^2
      //double y = x[1], jacy = 1.;
      //double y = 1.-x[1]*x[1], jacy = 2*x[1];
      //
      // t change of variable
      //const double pbeta1 = 2.4; const double pbeta2 = 3.2; double t = pow(1.-x[0],pbeta2)/pow(x[0],pbeta1); double jac = (pbeta1+x[0]*(pbeta2-pbeta1)) * pow(1.-x[0],pbeta2-1.) / pow(x[0],pbeta1+1.);
      //const double pbeta1 = 1.4; const double pbeta2 = 1.2; double t = pow(1.-x[0],pbeta2)/pow(x[0],pbeta1); double jac = (pbeta1+x[0]*(pbeta2-pbeta1)) * pow(1.-x[0],pbeta2-1.) / pow(x[0],pbeta1+1.);
      //double mult = 3; double u0 = 1./a0/b0+log(1-y); double u = x[0]*u0; double t1 = exp(-u); double t2 = exp(mult*u); double jac1 = t1*u0; double jac2 = mult*t2*u0;
      // next one used for the Higgs paper
      //double mult=3; double u0=1./a0/b0-lFQ; double xi0=exp(-u0); double u=x[0]*u0; double t1=exp(-u); double t2=exp(mult*u); double jac1=t1*u0; double jac2=mult*t2*u0; double y1=(1-xi0/t1)*(1.-x[1]*x[1]); double y2=(1-xi0/t2)*(1.-x[1]*x[1]); jac1*=(1-xi0/t1)*2*x[1]; jac2*=(1-xi0/t2)*2*x[1];
      // next one with improved change of variables
      double mult=3; double u0=1./a0/b0; double xi0=xiF*exp(-u0); double u=x[0]*u0; double t1=xiF*exp(-u); double t2=xiF*exp(mult*u); double jac1=t1*u0; double jac2=mult*t2*u0; double y1=(1-xi0/t1)*(1.-x[1]*x[1]); double y2=(1-xi0/t2)*(1.-x[1]*x[1]); jac1*=(1-xi0/t1)*2*x[1]; jac2*=(1-xi0/t2)*2*x[1];
      //double mult=3; double u0=1./a0/b0-lFQ; double u=x[0]*u0; double t1 = exp(-u); double t2 = exp(mult*u); double jac1 = t1*u0; double jac2 = mult*t2*u0; double y1 = (1-exp(-u0)/t1)*x[1]; double y2 = (1-exp(-u0)/t2)*x[1]; jac1*=(1-exp(-u0)/t1); jac2*=(1-exp(-u0)/t2);
      //cout << xi0 << endl;
      //
      //dcomplex result = jac * integrand2d(t, y, g, r, pr, a0*b0, par);
      //dcomplex result = jac1 * integrand2d(t1, y, g, r, pr, a0*b0, par) + jac2 * integrand2d(t2, y, g, r, pr, a0*b0, par);
      result = jac1 * integrand2d(t1, y1, g, r, pr, a0*b0, par) + jac2 * integrand2d(t2, y2, g, r, pr, a0*b0, par);
      //result *= jacy; // jacobian for y change of variable
      //cout << result << "  " << t2 << "  " << y2 << endl;
    } else if(whichFunction == 1) {
      // next one used for the Higgs paper
      //double u0=1./a0/b0-lFQ; double xi0=exp(-u0); double u=x[0]*u0; double t1=exp(-u); double jac1=t1*u0; double y1=(1-xi0/t1)*(1.-x[1]*x[1]); jac1*=(1-xi0/t1)*2*x[1];
      // next one with improved change of variables
      double u0=1./a0/b0; double xi0=xiF*exp(-u0); double u=x[0]*u0; double t1=xiF*exp(-u); double jac1=t1*u0; double y1=(1-xi0/t1)*(1.-x[1]*x[1]); jac1*=(1-xi0/t1)*2*x[1];
      result = jac1 * integrand2d(t1, y1, g, r, pr, a0*b0, par);
    } else if(whichFunction == 2) {
      // next one used for the Higgs paper
      //double mult=3; double u0=1./a0/b0-lFQ; double xi0=exp(-u0); double u=x[0]*u0; double t2=exp(mult*u); double jac2=mult*t2*u0; double y2=(1-xi0/t2)*(1.-x[1]*x[1]); jac2*=(1-xi0/t2)*2*x[1];
      // next one with improved change of variables
      double mult=3; double u0=1./a0/b0; double xi0=xiF*exp(-u0); double u=x[0]*u0; double t2=xiF*exp(mult*u); double jac2=mult*t2*u0; double y2=(1-xi0/t2)*(1.-x[1]*x[1]); jac2*=(1-xi0/t2)*2*x[1];
      result = jac2 * integrand2d(t2, y2, g, r, pr, a0*b0, par);
    }
    //
    res[0] = real(result);
    res[1] = imag(result);
    if(res[0]!=res[0]) res[0]=0;  // tappullo!!
    if(res[1]!=res[1]) res[1]=0;  // tappullo!!
    //cout << x[0] << "   " << x[1] << "   " << result << endl;
    return 0;
  }
  int integrandCUBA0(const int *ndim, const double x[], const int *ncomp, double res[], void *pars) {
    return integrandCUBAgeneral(ndim, x, ncomp, res, pars, 0);
  }
  int integrandCUBA1(const int *ndim, const double x[], const int *ncomp, double res[], void *pars) {
    return integrandCUBAgeneral(ndim, x, ncomp, res, pars, 1);
  }
  int integrandCUBA2(const int *ndim, const double x[], const int *ncomp, double res[], void *pars) {
    return integrandCUBAgeneral(ndim, x, ncomp, res, pars, 2);
  }
  dcomplex integrandGSL(double x, void *pars) {
    s_pr3c2d p = *(s_pr3c2d*)pars;
    dcomplex N  = p.c1;
    dcomplex g  = p.c2;
    dcomplex r  = p.c3;
    double a0 = p.d3;  // as(Q^2)
    double b0 = p.d4;  // beta0
    //
    //double xi0 = a0*b0;
    //if(real(r)>0 && real(r)>xi0) xi0 = real(r); // not sure if it makes sense...
    //
    //double xi = exp(-1./a0/b0) + (1.-x)/x;  double jac = 1./x/x;
    //double mult = 3; double t = (mult*x-1)/a0/b0; double xi = exp(t);  double jac = xi*mult/a0/b0;
    double mult = 3; double t = x/a0/b0; double xi1 = exp(-t); double xi2 = exp(mult*t); double jac1 = xi1/a0/b0; double jac2 = mult*xi2/a0/b0;
    //double mult = 3; double u=x*x; double t = u/a0/b0; double xi1 = exp(-t); double xi2 = exp(mult*t); double jac1 = xi1/a0/b0*2*x; double jac2 = mult*xi2/a0/b0*2*x;
    //double amin = 0.01*a0, amax = 2000*a0;  double a = amin + (amax-amin)*x;  double xi = exp((1/a-1/a0)/b0);  double jac = -xi/(-b0*a*a) * (amax-amin);
    //double Xmin = 0.1, Xmax = 80;  double X = Xmin + (Xmax-Xmin)*x;  double xi = exp((1-X)/X/xi0);  double jac = xi/X/X/xi0*(Xmax-Xmin);
    //double Xmin = 0.05, Xmax = 70;  double X1 = Xmin + (1-Xmin)*x;  double xi1 = exp((1-X1)/X1/xi0);  double jac1 = xi1/X1/X1/xi0*(1-Xmin); double X2 = 1 + (Xmax-1)*(1-x);  double xi2 = exp((1-X2)/X2/xi0);  double jac2 = xi2/X2/X2/xi0*(Xmax-1);
    //double Xmin = 0.1;  double X1 = Xmin + (1-Xmin)*x;  double xi1 = exp((1-X1)/X1/xi0);  double jac1 = xi1/X1/X1/xi0*(1-Xmin);  double X2 = x;  double xi2 = exp((1-1/X2)*X2/xi0);  double jac2 = xi2/xi0;
    //cout << x << "  " << xi << endl;
    //if(1+real(r)*log(xi1)<0) cout << "warning! xi1="<< xi1 << "  r="<< r << endl;
    //if(1+real(r)*log(xi2)<0) cout << "warning! xi2="<< xi2 << "  r="<< r << endl;
    //
    void *par = p.par;
    //dcomplex result = jac * integrand(xi, N, g, r, p.p, a0*b0, 0, par);
    dcomplex result = jac1 * integrand(xi1, N, g, r, p.p, a0*b0, 0, par) + jac2 * integrand(xi2, N, g, r, p.p, a0*b0, 0, par);
    //
    //cout << setw(15) << xi << setw(15) << x[1] << setw(35) << res << endl;
    return result;
  }
  double integrandGSLre(double x, void *p) {
    return real(integrandGSL(x,p));
  }
  double integrandGSLim(double x, void *p) {
    return imag(integrandGSL(x,p));
  }


  extern dcomplex gamma2sing(dcomplex N, double nf, bool var);
  dcomplex newDeltaFRes(dcomplex Delta_gammaLO_BatNLO, dcomplex Delta_gammaLOdot_BatNLO, dcomplex N, double as, double nf, process proc, int ordersub=0, void *pars=NULL) {
    hCoeff hC(proc, 1, pars);
    double h0 = hC.hk(0);
    //
    double b0 = beta0(nf);
    dcomplex gamma0 = gamma0sing(N,nf);
    dcomplex gamma1 = gamma1sing(N,nf);
    //dcomplex gamma2 = (useLLp ? 0 : gamma2sing(N,nf,RCvar) );
    dcomplex gamma2 = gamma2sing(N,nf,RCvar);
    //
    //gamma0 = gamma0(N,nf) - gamma0(N+1.,nf);
    //gamma0 = gamma0(N,nf) - 2.*gamma0(N+1.,nf) + gamma0(N+2.,nf);
    //gamma0 = gamma0(N,nf) - 3.*gamma0(N+1.,nf) + 3.*gamma0(N+2.,nf) - gamma0(N+3.,nf);
    //
    //*
    // Marco version
    dcomplex gammaLO_BatNLO    = Delta_gammaLO_BatNLO    + as*gamma0 + as*as*gamma1 + as*as*as*gamma2;
    dcomplex gammaLOdot_BatNLO = Delta_gammaLOdot_BatNLO +    gamma0 + 2.*as*gamma1 + 3.*as*as*gamma2;
    //
    //gammaLO_BatNLO   *= 2./(N+2.); //hack!!!
    //gammaLO_BatNLO /= N+1.; // hack! large-N damping
    //gammaLOdot_BatNLO /= N+1.; // hack! large-N damping
    if(RCmode == 1) gammaLOdot_BatNLO = gammaLO_BatNLO/as;
    //
    gammaLOdot_BatNLO *= -b0*as*as;  // convert deriv wrt as to deriv wrt t
    dcomplex r = -gammaLOdot_BatNLO / gammaLO_BatNLO;
    //
    s_pr3c2d intpars = { proc, N, gammaLO_BatNLO, r, as, b0, pars };
    dcomplex res;
    if(proc == ggH) {
      const int ncomp = 2;
      double cres[ncomp], error[ncomp], prob[ncomp];
      double epsrel=1e-2, epsabs=1.e-10;
      //if(proc == C2) epsrel /= 3;
      int last = 4;
      int verbose = 0;
      int nregions, neval, fail;
      int mineval = 0, maxeval = 5000000;
      //nregions=1; // to remove the warning
      int dim = 2;
      //int seed = 0;
      //
      // int NNN = 50;
      // for(int i=0; i<=NNN; i++) {
      // 	double xxx[2]={i/(NNN+0.),0.02}, xxres[2];
      // 	integrandCUBA(&dim, xxx, &ncomp, xxres, &intpars);
      // 	cout << setw(15) << xxx[0] << "  " << xxx[1] << "     " << setw(15) << xxres[0] << "  " << setw(15) << xxres[1] << endl;
      // }
      // exit(0);
      // cout << endl;
      //
      // first integral
      Cuhre(dim, ncomp, integrandCUBA1, &intpars, 1, epsrel, epsabs, verbose | last,       mineval, maxeval, 13, NULL, NULL, &nregions, &neval, &fail, cres, error, prob);
      //Vegas(dim, ncomp, integrandCUBA, &intpars, 1, epsrel, epsabs, verbose | last, seed, mineval, maxeval, 50000, 20000, 20000, 1, NULL, NULL, &neval, &fail, cres, error, prob);
      //Suave(dim, ncomp, integrandCUBA, &intpars, 1, epsrel, epsabs, verbose | last, seed, mineval, maxeval, 50000, 20000, 1, NULL, NULL, &nregions, &neval, &fail, cres, error, prob);
      if(fail == -1)  cout << "  ERROR: dimension out of range." << endl;
      else if(fail>0) cout << "  ERROR: precision not reached." << endl;
      //if(!verbose) cout << "Integral result: " << setw(12) << cres[0] << " +/- " << setw(12) << error[0] << "  (relative error " << setw(12) << error[0]/cres[0] <<")" << endl;
      //if(!verbose) cout << "Integral result: " << setw(12) << cres[1] << " +/- " << setw(12) << error[1] << "  (relative error " << setw(12) << error[1]/cres[1] <<")" << endl;
      res = cres[0] + I*cres[1];
      //cout << res << endl;
      //if(res != 1.) exit(0);
      //
      // second integral
      Cuhre(dim, ncomp, integrandCUBA2, &intpars, 1, epsrel, epsabs, verbose | last,       mineval, maxeval, 13, NULL, NULL, &nregions, &neval, &fail, cres, error, prob);
      //Vegas(dim, ncomp, integrandCUBA, &intpars, 1, epsrel, epsabs, verbose | last, seed, mineval, maxeval, 50000, 20000, 20000, 1, NULL, NULL, &neval, &fail, cres, error, prob);
      //Suave(dim, ncomp, integrandCUBA, &intpars, 1, epsrel, epsabs, verbose | last, seed, mineval, maxeval, 50000, 20000, 1, NULL, NULL, &nregions, &neval, &fail, cres, error, prob);
      if(fail == -1)  cout << "  ERROR: dimension out of range." << endl;
      else if(fail>0) cout << "  ERROR: precision not reached." << endl;
      //if(!verbose) cout << "Integral result: " << setw(12) << cres[0] << " +/- " << setw(12) << error[0] << "  (relative error " << setw(12) << error[0]/cres[0] <<")" << endl;
      //if(!verbose) cout << "Integral result: " << setw(12) << cres[1] << " +/- " << setw(12) << error[1] << "  (relative error " << setw(12) << error[1]/cres[1] <<")" << endl;
      res += cres[0] + I*cres[1];
      //cout << res << endl;
      //if(res != 1.) exit(0);
      //
    }
    else {
      gsl_error_handler_t *old_handler = gsl_set_error_handler (NULL);
      gsl_set_error_handler_off();
      int status;
      double prec = 1e-4;
      res = integration_gauss(integrandGSLre, 0, 1, prec, NULL, &intpars, &status)
	+ I*integration_gauss(integrandGSLim, 0, 1, prec, NULL, &intpars, &status);
      if(status) res = 0;
      gsl_set_error_handler (old_handler);
    }
    /*/
    // Richard version // not yet modified to account for the fact that Delta is now Delta2!!
    dcomplex FOgammaLO  = as*gamma0pl(N+1.,0);
    dcomplex gammaLO_BatNLO    = Delta_gammaLO_BatNLO    + FOgammaLO;
    dcomplex gammaLOdot_BatNLO = Delta_gammaLOdot_BatNLO + FOgammaLO/as;
    if(RCmode == 1) gammaLOdot_BatNLO = gammaLO_BatNLO/as;
    gammaLOdot_BatNLO *= -b0*as*as;  // convert deriv wrt as to deriv wrt t
    FOgammaLO -= as*gamma0sing(N,0);
    dcomplex FOgammaLOdot  = -b0*as*FOgammaLO;
    //gammaLOdot_BatNLO = 0;
    //FOgammaLOdot = 0;
    //
    dcomplex r   = -gammaLOdot_BatNLO / gammaLO_BatNLO;
    dcomplex rLO = -FOgammaLOdot / FOgammaLO;
    //
    s_pr2c2d intpars   = { proc,   gammaLO_BatNLO, r,   as, b0 };
    s_pr2c2d intparsLO = { proc, FOgammaLO,        rLO, as, b0 };
    dcomplex res;
    gsl_error_handler_t *old_handler = gsl_set_error_handler (NULL);
    gsl_set_error_handler_off();
    double prec = 1e-4;
    res  = integration_gauss(integrandGSLre, 0, 1, prec, NULL, &intpars)
      +  I*integration_gauss(integrandGSLim, 0, 1, prec, NULL, &intpars);
    res -= integration_gauss(integrandGSLre, 0, 1, prec, NULL, &intparsLO)
      +  I*integration_gauss(integrandGSLim, 0, 1, prec, NULL, &intparsLO);
    res += h0;
    gsl_set_error_handler (old_handler);
    //*/
    //
    //if(proc == gamma_qg) res += evolU(1.,gammaLO_BatNLO,r) * Dxi_Hqg(1.);
    //
    //if(useLLp) gamma2 = gamma2sing(N,nf,RCvar); // recompute it since it was set to zero before
    double T = (RCmode==1 ? 1 : 2);
    if(ordersub>0) res -= h0;
    if(proc == ggH) {
      double h10 = hC.hk(1);
      double h20 = hC.hk(2);
      double h11 = hC.hk(3);
      double h30 = hC.hk(4);
      double h21 = hC.hk(5);
      if(ordersub>1) res -= 2*h10*as*gamma0;
      if(ordersub>2) res -= as*as*(2*h20*gamma0*(gamma0-b0) + 2*h10*gamma1 + h11*gamma0*gamma0);
      if(ordersub>3) res -= as*as*as*(2*h30*gamma0*(gamma0-b0)*(gamma0-2*b0) + 2*h20*gamma1*(2.*gamma0-T*b0) + 2*h10*gamma2 + 2*h21*gamma0*gamma0*(gamma0-b0) + 2*h11*gamma1*gamma0);
    } else {
      double h1 = hC.hk(1);
      double h2 = hC.hk(2);
      double h3 = hC.hk(3);
      if(ordersub>1) res -= h1*as*gamma0;
      if(ordersub>2) res -= as*as*(h2*gamma0*(gamma0-b0) + h1*gamma1);
      if(ordersub>3) res -= as*as*as*(h3*gamma0*(gamma0-b0)*(gamma0-2*b0) + h2*gamma1*(2.*gamma0-T*b0) + h1*gamma2);
    }
    if(ordersub>4) cout << "Error, trying to subtract beyond the capability of HELL" << endl;
    //cout << setw(15) << real(gammaLO_BatNLO) << setw(16) << real(res) << endl;
    return res;
  }


  dcomplex DeltaCtest1Res(dcomplex Delta_gammaLO_BatNLO, dcomplex Delta_gammaLOdot_BatNLO, dcomplex N, double as, double nf, dcomplex Delta_gammaNLO, void *pars) {
    return newDeltaFRes(Delta_gammaLO_BatNLO, Delta_gammaLOdot_BatNLO, N, as, nf, Ctest1, 0);
  }
  dcomplex DeltaCtest2Res(dcomplex Delta_gammaLO_BatNLO, dcomplex Delta_gammaLOdot_BatNLO, dcomplex N, double as, double nf, dcomplex Delta_gammaNLO, void *pars) {
    return newDeltaFRes(Delta_gammaLO_BatNLO, Delta_gammaLOdot_BatNLO, N, as, nf, Ctest2, 0);
  }
  dcomplex newDeltaGammaResNLOqg(dcomplex Delta_gammaLO_BatNLO, dcomplex Delta_gammaLOdot_BatNLO, dcomplex N, double as, double nf, dcomplex Delta_gammaNLO, void *pars) {
    dcomplex res = newDeltaFRes(Delta_gammaLO_BatNLO, Delta_gammaLOdot_BatNLO, N, as, nf, gamma_qg, 2);
    return res * as*nf/3./M_PI;
  }

  // matching condition
  dcomplex DeltaKhgRes(dcomplex Delta_gammaLO_BatNLO, dcomplex Delta_gammaLOdot_BatNLO, dcomplex N, double as, double nf, dcomplex Delta_gammaNLO, void *pars) {
    dcomplex res = newDeltaFRes(Delta_gammaLO_BatNLO, Delta_gammaLOdot_BatNLO, N, as, nf, Khg, 2, pars);
    res *= as/3./M_PI;
    res += DeltaUqg(Delta_gammaLO_BatNLO, Delta_gammaLOdot_BatNLO, N, as, nf, Delta_gammaNLO)/nf;
    return res;
  }

  // massless DIS Coeff Functions
  dcomplex DeltaCLRes(dcomplex Delta_gammaLO_BatNLO, dcomplex Delta_gammaLOdot_BatNLO, dcomplex N, double as, double nf, dcomplex Delta_gammaNLO, void *pars) {
    dcomplex res = newDeltaFRes(Delta_gammaLO_BatNLO, Delta_gammaLOdot_BatNLO, N, as, nf, CL, 2);
    return res * as*nf/3./M_PI;
    //return res * as/3./M_PI;
  }
  dcomplex DeltaC2Res(dcomplex Delta_gammaLO_BatNLO, dcomplex Delta_gammaLOdot_BatNLO, dcomplex N, double as, double nf, dcomplex Delta_gammaNLO, void *pars) {
    dcomplex res = newDeltaFRes(Delta_gammaLO_BatNLO, Delta_gammaLOdot_BatNLO, N, as, nf, C2, 2);
    res *= as*nf/3./M_PI;
    //res *= as/3./M_PI;
    res -= DeltaUqg(Delta_gammaLO_BatNLO, Delta_gammaLOdot_BatNLO, N, as, nf, Delta_gammaNLO);
    return res;
  }

  // massive DIS Coeff Functions in Neutral Current
  dcomplex DeltaMC2Res(dcomplex Delta_gammaLO_BatNLO, dcomplex Delta_gammaLOdot_BatNLO, dcomplex N, double as, double nf, dcomplex Delta_gammaNLO, void *pars) {
    dcomplex res = newDeltaFRes(Delta_gammaLO_BatNLO, Delta_gammaLOdot_BatNLO, N, as, nf, MC2, 2, pars);
    Interpolator *interp = (Interpolator*) pars;
    double X = 1/(1+4*interp->Getxim());
    res *= pow(X,N)/(N*(1.-X)+1.);      // restores the theta function in N space
    return res * as/3./M_PI;
  }
  dcomplex DeltaMCLRes(dcomplex Delta_gammaLO_BatNLO, dcomplex Delta_gammaLOdot_BatNLO, dcomplex N, double as, double nf, dcomplex Delta_gammaNLO, void *pars) {
    dcomplex res = newDeltaFRes(Delta_gammaLO_BatNLO, Delta_gammaLOdot_BatNLO, N, as, nf, MCL, 2, pars);
    Interpolator *interp = (Interpolator*) pars;
    double X = 1/(1+4*interp->Getxim());
    res *= pow(X,N)/(N*(1.-X)+1.);      // restores the theta function
    return res * as/3./M_PI;
  }
  dcomplex DeltaMC2ResAxial(dcomplex Delta_gammaLO_BatNLO, dcomplex Delta_gammaLOdot_BatNLO, dcomplex N, double as, double nf, dcomplex Delta_gammaNLO, void *pars) {
    dcomplex res = newDeltaFRes(Delta_gammaLO_BatNLO, Delta_gammaLOdot_BatNLO, N, as, nf, MC2ax, 2, pars);
    Interpolator *interp = (Interpolator*) pars;
    double X = 1/(1+4*interp->Getxim());
    res *= pow(X,N)/(N*(1.-X)+1.);      // restores the theta function
    return res * as/3./M_PI;
  }
  dcomplex DeltaMCLResAxial(dcomplex Delta_gammaLO_BatNLO, dcomplex Delta_gammaLOdot_BatNLO, dcomplex N, double as, double nf, dcomplex Delta_gammaNLO, void *pars) {
    dcomplex res = newDeltaFRes(Delta_gammaLO_BatNLO, Delta_gammaLOdot_BatNLO, N, as, nf, MCLax, 2, pars);
    Interpolator *interp = (Interpolator*) pars;
    double X = 1/(1+4*interp->Getxim());
    res *= pow(X,N)/(N*(1.-X)+1.);      // restores the theta function
    return res * as/3./M_PI;
  }

  // massive DIS Coeff Functions in Charged Current
  dcomplex DeltaMC2ResCC(dcomplex Delta_gammaLO_BatNLO, dcomplex Delta_gammaLOdot_BatNLO, dcomplex N, double as, double nf, dcomplex Delta_gammaNLO, void *pars) {
    dcomplex res = newDeltaFRes(Delta_gammaLO_BatNLO, Delta_gammaLOdot_BatNLO, N, as, nf, MC2cc, 2, pars);
    res *= as/3./M_PI;
    res -= DeltaUqg(Delta_gammaLO_BatNLO, Delta_gammaLOdot_BatNLO, N, as, nf, Delta_gammaNLO)/nf/2.;
    Interpolator *interp = (Interpolator*) pars;
    double X = 1/(1+interp->Getxim());
    res *= pow(X,N)/(N*(1.-X)+1.);      // restores the theta function
    return res;
  }
  dcomplex DeltaMCLResCC(dcomplex Delta_gammaLO_BatNLO, dcomplex Delta_gammaLOdot_BatNLO, dcomplex N, double as, double nf, dcomplex Delta_gammaNLO, void *pars) {
    dcomplex res = newDeltaFRes(Delta_gammaLO_BatNLO, Delta_gammaLOdot_BatNLO, N, as, nf, MCLcc, 2, pars);
    res *= as/3./M_PI;
    Interpolator *interp = (Interpolator*) pars;
    double xim = interp->Getxim();
    res -= DeltaUqg(Delta_gammaLO_BatNLO, Delta_gammaLOdot_BatNLO, N, as, nf, Delta_gammaNLO)/nf/2. * xim/(1.+xim);
    double X = 1/(1+xim);
    res *= pow(X,N)/(N*(1.-X)+1.);      // restores the theta function
    return res;
  }
  dcomplex DeltaMC3ResCC(dcomplex Delta_gammaLO_BatNLO, dcomplex Delta_gammaLOdot_BatNLO, dcomplex N, double as, double nf, dcomplex Delta_gammaNLO, void *pars) {
    dcomplex res = newDeltaFRes(Delta_gammaLO_BatNLO, Delta_gammaLOdot_BatNLO, N, as, nf, MC3cc, 2, pars);
    res *= as/3./M_PI;
    Interpolator *interp = (Interpolator*) pars;
    double xim = interp->Getxim();
    res -= DeltaUqg(Delta_gammaLO_BatNLO, Delta_gammaLOdot_BatNLO, N, as, nf, Delta_gammaNLO)/nf/2. / (1.+xim);
    double X = 1/(1+xim);
    res *= pow(X,N)/(N*(1.-X)+1.);      // restores the theta function
    return res;
  }


  // ggH inclusive XS
  dcomplex DeltaHggRes(dcomplex Delta_gammaLO_BatNLO, dcomplex Delta_gammaLOdot_BatNLO, dcomplex N, double as, double nf, dcomplex Delta_gammaNLO, void *pars) {
    dcomplex res = newDeltaFRes(Delta_gammaLO_BatNLO, Delta_gammaLOdot_BatNLO, N, as, nf, ggH, 4, pars);
    return res;
  }
  dcomplex DeltaHggAuxRes(dcomplex Delta_gammaLO_BatNLO, dcomplex Delta_gammaLOdot_BatNLO, dcomplex N, double as, double nf, dcomplex Delta_gammaNLO, void *pars) {
    dcomplex res = newDeltaFRes(Delta_gammaLO_BatNLO, Delta_gammaLOdot_BatNLO, N, as, nf, ggHaux, 4, pars);
    return res;
  }


};

