/* -----------------------------------------

   _  _ ____ _    _    
   |__| |___ |    |    
   |  | |___ |___ |___ 
   
   HELL: High-Energy Large Logarithms

   Author: Marco Bonvini

   Computes Delta P_res = P_res - P_FixedOrder
   from a file containing parameters

   ----------------------------------------- */

#include <iomanip>
#include <iostream>
#include <fstream>
#include <string>
#include <sstream>
#include <cmath>
#include <cstdlib>

#include "hell.hh"
#include "Pqg.hh"
#include "CoefficientFunctions.hh"
#include "CoefficientFunctionsNew.hh"
#include "CoefficientFunctionsOffShellInterpolator.hh"
#include "hHiggs_new_var.hh"
#include "integration.hh"


using namespace std;



namespace HELL {

  bool useLLp = false;
  void SetLLpMode(bool use) {
    useLLp = use;
  }

  /*
dcomplex dpsi(dcomplex Z, int M=1){
  int K1, K2;
  dcomplex SUB = 0. , SUBM;
  dcomplex ZZ = Z;
  if(fabs(imag(ZZ))<10.) { // if too close to the real axis...
  label1:
    SUBM = -1./ZZ;
    for(K1=1; K1<=M; K1++) {
      SUBM = - SUBM * (double)K1 / ZZ;
    }
    if(real(ZZ)<10.) { // ...use recurrence relation to push real(z) large enough
      SUB = SUB + SUBM;
      ZZ = ZZ + 1.;
      goto label1;
    }
  }
  // Expansion coefficients for the first derivative
  double A1 =  1.;
  double A2 =  1./2.;
  double A3 =  1./6.;
  double A4 = -1./30.;
  double A5 =  1./42.;
  double A6 = -1./30.;
  double A7 =  5./66.;
  // Expansion coefficients for the higher derivatives
  if(M>1) {
    for(K2=2;K2<=M;K2++){
      A1 = A1 * (1.*K2-1.);
      A2 = A2 *  1.*K2;
      A3 = A3 * (1.*K2+1.);
      A4 = A4 * (1.*K2+3.);
      A5 = A5 * (1.*K2+5.);
      A6 = A6 * (1.*K2+7.);
      A7 = A7 * (1.*K2+9.);
    }
  }
  dcomplex RZ = 1./ ZZ;
  dcomplex DZ = RZ * RZ;
  // SUB + asympt expansion (Abramowitz, Stengun, 6.4.11)
  return SUB + pow(-1.,M+1) * pow(RZ,M) * ( A1 + RZ * (A2 + RZ * (A3 + DZ * (A4 + DZ * (A5 + DZ * (A6 + A7 * DZ ))))) );
}
  double a11()       { return CA/M_PI; }
  double a10(int nf) { return -(11.*CA + 2.*nf*(1.-2.*CF/CA))/12./M_PI; }
  double a21(int nf) { return nf*(26*CF-23*CA)/36/M_PI/M_PI; }
  dcomplex nfa_sx_gammaAP_LO(dcomplex N, double as, int nf) {
    double a1 = as*a11();
    double a0 = as*a10(nf);
    return a1/N + a0 + 2*(a1+a0)*(1./(N+1.)-1.); // preserves momentum conservation, accurate at NLL, constant at large N
  }
  dcomplex nfa_sx_gammaAP_LO_prime(dcomplex N, double as, int nf) {
    double a1 = as*a11();
    double a0 = as*a10(nf);
    return -a1/N/N - 2*(a1+a0)/(N+1.)/(N+1.);
  }
  dcomplex gamma_DLNLO_order_as3(dcomplex N, double as, int nf) {
    double A = ( CA*CA*(-74./27. +11./12.*ZETA2 +5./2.*ZETA3) +nf*((4*CA+7*CF)/27. +(CA-2*CF)*ZETA2/6.) )/M_PI/M_PI;
    double chi02 = a10(nf)*a11();
    double chi11 = a21(nf);
    return -as*as*as*nfa_sx_gammaAP_LO_prime(N, 1, nf) * ( A + chi11/(N+1.) + chi02/pow(N+1.,2) - (A+chi11/2.+chi02/4.)*fmom(N) + CA/M_PI*(dpsi(N+1.)-dpsi(1))*nfa_sx_gammaAP_LO(N, 1, nf));
  }
  const double c0 = CA/M_PI*4*log(2.);
  const double k0 = CA/M_PI*28.*ZETA3;
  dcomplex gamma_BatemanColl_minus_dcNLO_momcons_order_as3(dcomplex N, double as, int nf) {
    double b0 = beta0(nf);
    double res = b0*k0/16.;
    return as*as*as*b0* res * (1./N - fmom(N));
  }
  dcomplex gamma_BatemanColl_minus_dcLO_momcons_order_as2 (dcomplex N, double as, int nf) {
    return as*as*beta0(nf)* (3*k0/32.-c0) * (1./N - fmom(N));
  }
  dcomplex DeltaGammaLO(dcomplex Delta_gamma, dcomplex Delta_gammadot, dcomplex N, double as, double nf, dcomplex Delta_gammaNLO, void *pars) {
    //Delta_gamma += gamma_BatemanColl_minus_dcLO_momcons_order_as2(N,as,nf);
    return Delta_gamma;
  }
  dcomplex DeltaGammaNLO(dcomplex Delta_gamma, dcomplex Delta_gammadot, dcomplex N, double as, double nf, dcomplex Delta_gammaNLO, void *pars) {
    //Delta_gamma += gamma_DLNLO_order_as3(N,as,nf) + gamma_BatemanColl_minus_dcNLO_momcons_order_as3(N,as,nf);
    return Delta_gamma;
  }
  */
  dcomplex DeltaGamma(dcomplex Delta_gamma, dcomplex Delta_gammadot, dcomplex N, double as, double nf, dcomplex Delta_gammaNLO, void *pars) {
    return Delta_gamma;
  }






  // **************************
  // ***   fixed nf part    ***
  // **************************


  HELLnf::~HELLnf() {
    // should write the various delete...
  }


  void HELLnf::Init(string datapath) {
    _datapath = datapath;
    //string pyexec = "python "+datapath+"/gen_info.py "+datapath;
    //system(pyexec.c_str());
    if(_order==0) {
      ostringstream filename;
      filename << datapath << "/LL_nf" << _nf << ".info";
      ifstream info(filename.str().c_str());
      double astmp;
      if(!info.good()) {
	cout << "\033[0;31m" << "HELL: Error reading info file " << filename.str() << "\033[0m" << endl;
	exit(0);
      }
      while(info.good()) {
	info >> astmp;
	_alphas.push_back(astmp);
      }
      _alphas.pop_back();
    } else
      if (_order == 1) {
      ostringstream filename;
      filename << datapath << "/NLL_nf" << _nf << ".info";
      ifstream info(filename.str().c_str());
      double astmp;
      if(!info.good()) {
	cout << "\033[0;31m" << "HELL: Error reading info file " << filename.str() << "\033[0m" << endl;
	exit(0);
      }
      while(info.good()) {
	info >> astmp;
	_alphas.push_back(astmp);
      }
      _alphas.pop_back();
    }
  }

  void HELLnf::GetAvailableAlphas(vector<double> &as1) {
    as1.resize(_alphas.size());
    as1 = _alphas;
    return;
  }


  int HELLnf::alphas_interpolation(double as, vector<double> vas, double &factor) {
    int k = 0;
    if(as < vas[0] || as > vas[vas.size()-1]) {
      cout << "ERROR: alpha_s=" << as << " out of interpolation range [" << vas[0] << ", " << vas[vas.size()-1] << "]" << endl;
      exit(22);
    }
    for(unsigned int i=1; i<vas.size(); i++) {
      if(as <= vas[i]) break;
      k++;
    }
    //cout << alphas[k] << " < " << as << " < " << alphas[k+1] << endl;
    factor = (as-vas[k]) / (vas[k+1]-vas[k]);
    return k;
  }

  string sas(double as) {
    ostringstream os;
    if     (as<0.01) os << "000" << int(1000*as);
    else if(as<0.1 ) os << "00"  << int(1000*as);
    else if(as<1.  ) os << "0"   << int(1000*as);
    else os << int(1000*as);
    return os.str();
  }

  void HELLnf::ReadTable(int k) {
    ostringstream filename;
    string sord = (_order == 1 ? "NLL" : "LL");
    vector<int> kk;
    for(int j=0; j<2; j++) {
      itHTP = HTP.find(k+j);
      if(itHTP == HTP.end()) kk.push_back(k+j);
    }
    for(unsigned int i=0; i<kk.size(); i++) {
      filename << _datapath << "/N-space/gamma" << sord << "_nf" << _nf << "_alphas" << sas(_alphas[kk[i]]) << ".table";
      HTP[kk[i]] = new HellTableP();
      HTP[kk[i]]->readTable(filename.str());
      HTP[kk[i]]->setVariation(_RCvar);
      if(_order==1) HTP[kk[i]]->setMellinFunction(DeltaGamma);
      else          HTP[kk[i]]->setMellinFunction(DeltaGamma);
      filename.str("");  filename.clear();
    }
  }
  void HELLnf::ReadTableC(int k, string id, double d) {
    ostringstream filename;
    vector<int> kk;
    //
    for(int j=0; j<1; j++) {
      itHTP = HTP2[id][d].find(k+j);
      if(itHTP == HTP2[id][d].end()) kk.push_back(k+j); // if it's a new value of alphas for the given id and d, then recompute it
    }
    //
    for(unsigned int i=0; i<kk.size(); i++) {
      if(useLLp) filename << _datapath << "/N-space/gammaLLp_nf" << _nf << "_alphas" << sas(_alphas[kk[i]]) << ".table";
      else       filename << _datapath << "/N-space/gammaNLL_nf" << _nf << "_alphas" << sas(_alphas[kk[i]]) << ".table";
      HTP2[id][d][kk[i]] = new HellTableP();
      HTP2[id][d][kk[i]]->readTable(filename.str());
      filename.str("");  filename.clear();
      //
      if(useLLp) filename << _datapath << "/N-space/gammaLLpdot_nf" << _nf << "_alphas" << sas(_alphas[kk[i]]) << ".table";
      else       filename << _datapath << "/N-space/gammaNLLdot_nf" << _nf << "_alphas" << sas(_alphas[kk[i]]) << ".table";
      HTP2[id][d][kk[i]]->readDotTable(filename.str());
      filename.str("");  filename.clear();
      //
      //filename << _datapath << "/N-space/gammaNLL_nf" << _nf << "_alphas" << sas(_alphas[kk[i]]) << ".table";
      //if(id == "P+Uqg") HTP2[id][d][kk[i]]->readNLOTable(filename.str());
      //filename.str("");  filename.clear();
      //
      HTP2[id][d][kk[i]]->setVariation(_RCvar);
      //
      if(id == "PLLp")   HTP2[id][d][kk[i]]->setMellinFunction(Delta1Gamma);
      if(id == "PLLpd")  HTP2[id][d][kk[i]]->setMellinFunction(Delta1GammaDot);
      if(id == "P0")     HTP2[id][d][kk[i]]->setMellinFunction(Gamma0);
      if(id == "P0sq")   HTP2[id][d][kk[i]]->setMellinFunction(Gamma0sq);
      if(id == "P0cb")   HTP2[id][d][kk[i]]->setMellinFunction(Gamma0cb);
      if(id == "P1")     HTP2[id][d][kk[i]]->setMellinFunction(Gamma1);
      if(id == "P0P1")   HTP2[id][d][kk[i]]->setMellinFunction(Gamma0Gamma1);
      if(id == "P2")     HTP2[id][d][kk[i]]->setMellinFunction(Gamma2);
      if(id == "Pqg")    HTP2[id][d][kk[i]]->setMellinFunction(DeltaGammaResNLOqg);
      //if(id == "Pqg")    HTP2[id][d][kk[i]]->setMellinFunction(newDeltaGammaResNLOqg);
      //if(id == "P+Uqg")  HTP2[id][d][kk[i]]->setMellinFunction(DeltaGammaPlusUqg);
      if(id == "U")      HTP2[id][d][kk[i]]->setMellinFunction(DeltaUxi, &d);
      if(id == "Uqg")    HTP2[id][d][kk[i]]->setMellinFunction(DeltaUqg);
      if(id == "logR")   HTP2[id][d][kk[i]]->setMellinFunction(DeltaLogR);
      if(id == "C2")     HTP2[id][d][kk[i]]->setMellinFunction(DeltaC2Res);
      if(id == "CL")     HTP2[id][d][kk[i]]->setMellinFunction(DeltaCLRes);
      if(id == "oldCL")  HTP2[id][d][kk[i]]->setMellinFunction(oldDeltaCLRes);
      if(id == "oldC2")  HTP2[id][d][kk[i]]->setMellinFunction(oldDeltaC2Res);
      if(id == "Khg")    HTP2[id][d][kk[i]]->setMellinFunction(DeltaKhgRes, &d);
      if(id == "MC2")    HTP2[id][d][kk[i]]->setMellinFunction(DeltaMC2Res, Interp[d]);
      if(id == "MCL")    HTP2[id][d][kk[i]]->setMellinFunction(DeltaMCLRes, Interp[d]);
      if(id == "MC2ax")  HTP2[id][d][kk[i]]->setMellinFunction(DeltaMC2ResAxial, Interp[d]);
      if(id == "MCLax")  HTP2[id][d][kk[i]]->setMellinFunction(DeltaMCLResAxial, Interp[d]);
      if(id == "MC2CC")  HTP2[id][d][kk[i]]->setMellinFunction(DeltaMC2ResCC, Interp[d]);
      if(id == "MCLCC")  HTP2[id][d][kk[i]]->setMellinFunction(DeltaMCLResCC, Interp[d]);
      if(id == "MC3CC")  HTP2[id][d][kk[i]]->setMellinFunction(DeltaMC3ResCC, Interp[d]);
      if(id == "ggH")    HTP2[id][d][kk[i]]->setMellinFunction(DeltaHggRes, ggHsx[d]);
      if(id == "ggHaux") HTP2[id][d][kk[i]]->setMellinFunction(DeltaHggAuxRes, ggHsx[d]);
    }
  }
  void HELLnf::ReadTablex(int k) {
    itxTU = xTU[_RCvar][_RCmode].find(k);
    if(itxTU == xTU[_RCvar][_RCmode].end()) { // if it's a new value of alphas, then recompute it
      ostringstream filename;
      filename << _datapath << "/x-space-aux" << (useLLp ? "-LLp" : "") << "/xtableU";
      if(_RCvar ==1) filename << "var1";
      if(_RCmode==1) filename << "var2";
      filename << "_NLL_nf" << _nf << "_alphas" << sas(_alphas[k]) << ".table";
      xTU[_RCvar][_RCmode][k] = new xTableU(filename.str());
    }
  }
  double HELLnf::xDeltaCRes(string id, double x, double as, double nf, xTableU *tU, void *pars) {
    double res = 0;
    if(id == "C2")     res = DeltaC2Res (x, as, nf, tU, _muFrat, pars);
    if(id == "CL")     res = DeltaCLRes (x, as, nf, tU, _muFrat, pars);
    if(id == "Khg")    res = DeltaKhgRes(x, as, nf, tU, _muFrat, pars);
    if(id == "MC2")    res = DeltaMC2Res(x, as, nf, tU, _muFrat, pars);
    if(id == "MCL")    res = DeltaMCLRes(x, as, nf, tU, _muFrat, pars);
    if(id == "MC2ax")  res = DeltaMC2ResAxial(x, as, nf, tU, _muFrat, pars);
    if(id == "MCLax")  res = DeltaMCLResAxial(x, as, nf, tU, _muFrat, pars);
    if(id == "MC2CC")  res = DeltaMC2ResCC(x, as, nf, tU, _muFrat, pars);
    if(id == "MCLCC")  res = DeltaMCLResCC(x, as, nf, tU, _muFrat, pars);
    if(id == "MC3CC")  res = DeltaMC3ResCC(x, as, nf, tU, _muFrat, pars);
    if(id == "ggHaux") res = DeltaHggAuxRes(x, as, nf, tU, _muFrat, pars);
    if(id == "DY")     res = DeltaCDYRes(x, as, nf, tU, _muFrat, pars);
    return res;
  }

  // Delta P_+
  double HELLnf::xdeltaPplus(double as, double x) {
    double factor;
    int k = alphas_interpolation(as, _alphas, factor);
    ReadTable(k);
    double dPf_k1 = HTP[k  ]->eval(x);
    double dPf_k2 = HTP[k+1]->eval(x);
    double res = dPf_k1 + factor * (dPf_k2 - dPf_k1);
    return res * pow(1-x,_damping) * pow(1-sqrt(x),_dampingsqrt);
  }
  // Delta P_qg
  double HELLnf::xdeltaPqg(double as, double x) {
    //return xdeltaC(as, x, "P+Uqg"); // hack to get Pgg from the RC rotation
    return xdeltaC(as, x, "Pqg");
  }
  double HELLnf::xdeltaPpUqg(double as, double x) {
    return xdeltaC(as, x, "P+Uqg");
  }
  // Delta C
  double HELLnf::xdeltaC(double as, double x, string id, double mQ, double xi) {
    if(_order==0) return 0;
    // check if there was previously the same value of m/Q
    if(mQ>0) {
      itInterp = Interp.find(mQ);
      if(itInterp == Interp.end()) {
	//for(itInterp=Interp.begin(); itInterp!=Interp.end(); itInterp++) {
	//  //delete itInterp->second;
	//}
	Interp[mQ] = new Interpolator(mQ*mQ); // if it's a new value of m/Q, then recompute it
      }
    }
    //
    double factor;
    int k = alphas_interpolation(as, _alphas, factor);
    int k0 = -1;
    if(fabs(factor  )<1e-5) { k0 = k;  }
    if(fabs(factor-1)<1e-5) { k0 = k+1;}
    //cout << k0 << "  " << k << "   " << setprecision(20) << factor << endl; 
    double dCf[2]={0,0};
    if(_evaluateColdway || id=="PLLp" || id=="PLLpd" || id=="P0" || id=="P0sq" || id=="P0cb" || id=="P1" || id=="P0P1" || id=="P2" || id=="Pqg" || id=="Uqg" || id=="U" || id=="logR" || id=="oldCL" || id=="oldC2" || id=="ggH") {
      double d = 0;
      if(id == "U") d = xi;
      if(id=="Khg" || id=="MC2" || id=="MCL" || id=="MC2ax" || id=="MCLax" || id=="MC2CC" || id=="MCLCC" || id=="MC3CC") d = mQ;
      if(id == "ggH" || id=="ggHaux") d = _muFrat;
      //
      if(k0>=0) {
	ReadTableC(k0, id, d);
	dCf[0] = HTP2[id][d][k0]->eval(x);
	dCf[1] = dCf[0];
      } else {
	ReadTableC(k  , id, d);
	ReadTableC(k+1, id, d);
	//HTP2[id][mQ][k]->setErrorEstimation(true);
	dCf[0] = HTP2[id][d][k  ]->eval(x);
	dCf[1] = HTP2[id][d][k+1]->eval(x);
	//cout << dPf_k1 << "   " << dPf_k2 << endl;
	//cout << "error="<<HTP2[id][d][k]->getError() << endl;
      }
    } else {
      void *par=NULL;
      //ReadTablex(k);
      if(id=="Khg") par = &mQ;
      if(id=="MC2" || id=="MCL" || id=="MC2ax" || id=="MCLax" || id=="MC2CC" || id=="MCLCC" || id=="MC3CC") par = Interp[mQ];
      if(id=="ggHaux") par = ggHsx[_muFrat];
      if(k0>=0) {
	ReadTablex(k0);
	//cout << _alphas[k0] << endl;
	dCf[0] = xDeltaCRes(id, x, _alphas[k0], _nf, xTU[_RCvar][_RCmode][k0], par);
	dCf[1] = dCf[0];
      } else {
	ReadTablex(k);
	ReadTablex(k+1);
	dCf[0] = xDeltaCRes(id, x, _alphas[k  ], _nf, xTU[_RCvar][_RCmode][k  ], par);
	dCf[1] = xDeltaCRes(id, x, _alphas[k+1], _nf, xTU[_RCvar][_RCmode][k+1], par);
      }
    }
    /*
    if(mQ>0) {
      itInterp = Interp.find(mQ);
      if(itInterp != Interp.end()) {
	cout << "ciao" << endl;
	delete Interp[mQ];
      }
    }
    */
    double res = dCf[0] + factor * (dCf[1] - dCf[0]);
    if(res != res) res = 0;  //// tappullo!!!! to be fixed
    return res * pow(1-x,_damping) * pow(1-sqrt(x),_dampingsqrt);
  }


  struct intvars {
    HELLnf* h;
    double as;
  };
  double gMCintegrand(double t, void* p) {
    intvars vars = *(intvars *) p;
    HELLnf *hell = vars.h;
    double x = exp(-t);
    return x*( hell->xdeltaPplus(vars.as,x) + (1-CF/CA) * hell->xdeltaPqg(vars.as,x));
  }
  double HELLnf::gMC(double as) {
    itMC = MC.find(as);
    if (itMC == MC.end()) {
      double prec = 1e-4;
      intvars p = { this, as };
      double tmax = log(1.e6);
      //gsl_error_handler_t *old_handler = gsl_set_error_handler (NULL);
      //gsl_set_error_handler_off();
      MC[as] = integration_gauss(gMCintegrand, 0, tmax, prec, NULL, &p);
      //gsl_set_error_handler (old_handler);
    }
    return MC[as];
  }


  // computes d(1), where d(N) is the Mellin of the damping function (1-x)^k (1-sqrt(x))^j
  double d1(int k, int j) {
    double res = 0;
    for(int i=0; i<=k; i++) {
      res += gsl_sf_choose(k,i) * pow(-1.,i) /(2.+i) /gsl_sf_choose(4.+2*i+j,j);
    }
    return res;
  }


  // Splitting Function Matrix
  sqmatrix<double> HELLnf::DeltaPLL(double x, double as) {
    double dGgg = xdeltaPplus(as, x)/x;
    if(_momcons) {
      dGgg -= pow(1-x,_damping)*pow(1-sqrt(x),_dampingsqrt)/d1(_damping,_dampingsqrt) * gMC(as);
    }
    // Both of the following options are equally acceptable
    return sqmatrix<double>(dGgg, CF/CA*dGgg, 0., 0.);
    //return sqmatrix<double>(dGgg, 0., 0., 0.);
  }
  sqmatrix<double> HELLnf::DeltaPNLL(double x, double as) {
    double dGp  = xdeltaPplus(as, x)/x;
    double dGqg = xdeltaPqg(as, x)/x;
    double dGgg = dGp - CF/CA*dGqg;
    if(_momcons) {
      dGgg -= pow(1-x,_damping)*pow(1-sqrt(x),_dampingsqrt)/d1(_damping,_dampingsqrt) * gMC(as);
    }
    return sqmatrix<double>(dGgg, CF/CA*dGgg, dGqg, CF/CA*dGqg);
  }
  sqmatrix<double> HELLnf::DeltaP(double as, double x) {
    if(_order==0) {
      return DeltaPLL(x, as);
    } else if(_order==1) {
      return DeltaPNLL(x, as);
    }
    return sqmatrix<double>(0,0,0,0);
  }
  double HELLnf::deltaPplus(double as, double x) {
    return xdeltaPplus(as, x) /x;
  }
  double HELLnf::deltaPLLp(double as, double x) {
    return xdeltaC(as, x, "PLLp") /x;
  }
  double HELLnf::deltaPLLpd(double as, double x) {
    return xdeltaC(as, x, "PLLpd") /x;
  }
  double HELLnf::deltaP0(double as, double x) {
    return xdeltaC(as, x, "P0") /x;
  }
  double HELLnf::deltaP0sq(double as, double x) {
    return xdeltaC(as, x, "P0sq") /x;
  }
  double HELLnf::deltaP0cb(double as, double x) {
    return xdeltaC(as, x, "P0cb") /x;
  }
  double HELLnf::deltaP1(double as, double x) {
    return xdeltaC(as, x, "P1") /x;
  }
  double HELLnf::deltaP0P1(double as, double x) {
    return xdeltaC(as, x, "P0P1") /x;
  }
  double HELLnf::deltaP2(double as, double x) {
    return xdeltaC(as, x, "P2") /x;
  }
  double HELLnf::deltaLogR(double as, double x) {
    return xdeltaC(as, x, "logR") /x;
  }
  double HELLnf::deltaU(double as, double x, double xi) {
    return xdeltaC(as, x, "U", 0, xi) /x;
  }
  double HELLnf::deltaUqg(double as, double x) {
    return xdeltaC(as, x, "Uqg") /x;
  }
  // matching condition
  double HELLnf::deltaKhg(double as, double x, double m_Q_ratio) {
    return xdeltaC(as, x, "Khg", m_Q_ratio) /x;
  }
  //
  // Delta Coefficient function
  // massless case
  double HELLnf::deltaC2g(double as, double x) {
    //sqmatrix<double> dP = DeltaP(as,x,NLO);
    return xdeltaC(as, x, "C2") /x;// - dP.qg()/dP.gg();
  }
  double HELLnf::deltaCLg(double as, double x) {
    return xdeltaC(as, x, "CL") /x;
  }
  double HELLnf::deltaC2gold(double as, double x) {
    return xdeltaC(as, x, "oldC2") /x;
  }
  double HELLnf::deltaCLgold(double as, double x) {
    return xdeltaC(as, x, "oldCL") /x;
  }
  // massive case, neutral current
  double HELLnf::deltaMC2g(double as, double x, double m_Q_ratio) {
    if(4*m_Q_ratio*x/(1-x)>1) return 0;
    return xdeltaC(as, x, "MC2", m_Q_ratio) /x;
  }
  double HELLnf::deltaMCLg(double as, double x, double m_Q_ratio) {
    if(4*m_Q_ratio*x/(1-x)>1) return 0;
    return xdeltaC(as, x, "MCL", m_Q_ratio) /x;
  }
  double HELLnf::deltaMC2axg(double as, double x, double m_Q_ratio) {
    if(4*m_Q_ratio*x/(1-x)>1) return 0;
    return xdeltaC(as, x, "MC2ax", m_Q_ratio) /x;
  }
  double HELLnf::deltaMCLaxg(double as, double x, double m_Q_ratio) {
    if(4*m_Q_ratio*x/(1-x)>1) return 0;
    return xdeltaC(as, x, "MCLax", m_Q_ratio) /x;
  }
  // massive case, charged current
  double HELLnf::deltaMC2CCg(double as, double x, double m_Q_ratio) {
    if(m_Q_ratio*x/(1-x)>1) return 0;
    return xdeltaC(as, x, "MC2CC", m_Q_ratio) /x;
  }
  double HELLnf::deltaMCLCCg(double as, double x, double m_Q_ratio) {
    if(m_Q_ratio*x/(1-x)>1) return 0;
    return xdeltaC(as, x, "MCLCC", m_Q_ratio) /x;
  }
  double HELLnf::deltaMC3CCg(double as, double x, double m_Q_ratio) {
    if(m_Q_ratio*x/(1-x)>1) return 0;
    return xdeltaC(as, x, "MC3CC", m_Q_ratio) /x;
  }

  // Higgs coefficient function
  double HELLnf::deltaHgg(double as, double x, double mH, double mt, double mb, double mc) {
    //if(ggHsx) delete ggHsx;
    //ggHsx = new HiggsSmallX(mH, mt, mb, mc, _muFrat);
    if(!ggHsx[_muFrat]) {
      ggHsx[_muFrat] = new HiggsSmallX(mH, mt, mb, mc, _muFrat);
      ggHsx[_muFrat]->initCoeffs(3);
    }
    return xdeltaC(as, x, "ggH") /x;
  }
  double HELLnf::deltaHggAux(double as, double x, double mH, double mt, double mb, double mc) {
    //if(ggHsx) delete ggHsx;
    //ggHsx = new HiggsSmallX(mH, mt, mb, mc, _muFrat);
    if(!ggHsx[_muFrat]) {
      ggHsx[_muFrat] = new HiggsSmallX(mH, mt, mb, mc, _muFrat);
      ggHsx[_muFrat]->initCoeffs(3);
    }
    return xdeltaC(as, x, "ggHaux") /x;
  }

  // Drell-Yan coefficient function Cqg
  double HELLnf::deltaCDYqg(double as, double x) {
    return xdeltaC(as, x, "DY") /x;
  }


};

